/*******************************************************************************
* DISCLAIMER
* This software is supplied by Renesas Electronics Corporation and is only
* intended for use with Renesas products. No other uses are authorized. This
* software is owned by Renesas Electronics Corporation and is protected under
* all applicable laws, including copyright laws.
* THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES REGARDING
* THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT
* LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
* AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED.
* TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS
* ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES SHALL BE LIABLE
* FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES FOR
* ANY REASON RELATED TO THIS SOFTWARE, EVEN IF RENESAS OR ITS AFFILIATES HAVE
* BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
* Renesas reserves the right, without notice, to make changes to this software
* and to discontinue the availability of this software. By using this software,
* you agree to the additional terms and conditions found by accessing the
* following link:
* http://www.renesas.com/disclaimer
* Copyright (C) 2018 Renesas Electronics Corporation. All rights reserved.
*******************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

#include <r_bsp.h>
#include "r_bcd_main.h"
#include "r_opencv.h"
#include "ff.h"
//#include <stdio.h>

static uint32_t   	train_model_ready;
static uint16_t 	total_faces;
static const char*  model_filename_lbph = "fml";
static const char*  model_filename_eigen = "fme";
static const char*  model_filename_fisher = "fmf";
static const char*  record_number_filename = "record.bin";
#ifdef __cplusplus
}
#endif

#ifdef FACE_RECOGNITION_ENABLE
#include <opencv.hpp>
#include <opencv2/face.hpp>

using namespace cv;
using namespace std;
using namespace face;

Ptr<FaceRecognizer> LBPHRecog;
Ptr<FaceRecognizer> EigenRecog;
Ptr<FaceRecognizer> FisherRecog;
vector<int> labels;
vector<Mat> trainPics;
uint8_t algorithmSelected;

#ifdef __cplusplus
extern "C" {
#endif

FIL fp __attribute__ ((section ("BSS_SDHINC")));


FILE* fopen(const char *__restrict _name, const char *__restrict _type)
{
	if (_type == "wt")
	{
		if (f_open((FIL*)&fp, _name, FA_CREATE_ALWAYS | FA_WRITE) == FR_OK)
			return (FILE*)&fp;
		else
			return NULL;
	}
	else if (_type == "rt")
	{
		if (f_open((FIL*)&fp, _name, FA_READ) == FR_OK)
			return (FILE*)&fp;
		else
			return NULL;
	}

	return NULL;
}

int fclose(FILE *_f)
{
	if (f_close((FIL*)_f) == FR_OK)
		return 0;
	else
		return -1;
}

int fputs(const char *__restrict _s, FILE *__restrict _f)
{
	return f_puts((const TCHAR*)(_s), (FIL*)_f);
}

char * fgets(char *__restrict _s, int _l, FILE *__restrict _f)
{
	return f_gets((TCHAR*)(_s), _l, (FIL*)_f);
}

int	fseek(FILE *_f, long _o, int _w)
{
	if (_w == SEEK_END)
	{
		return f_lseek((FIL*)_f, 0xFFFFFFFF);
	}

	return 0;
}

long ftell(FILE *_f)
{
	return f_tell((FIL*)_f);
}

void rewind(FILE *_f)
{
	f_rewind((FIL*)_f);
}

int cv_face_model_delete(void)
{
	FRESULT ret0 = FR_OK;
	FRESULT ret1 = FR_OK;

	ret0 = f_unlink(model_filename_lbph);
//	ret0 = f_unlink(model_filename_eigen);
//	ret0 = f_unlink(model_filename_fisher);
	ret1 = f_unlink(record_number_filename);
	train_model_ready = 0;
	total_faces = 0;

	if (ret0 != FR_OK || ret1 != FR_OK)
		return -1;
	else
		return 0;
}

static int save_face_number(uint16_t face_id)
{
	if (f_open((FIL*)&fp, record_number_filename, FA_CREATE_ALWAYS) == FR_OK)
	{
		if (f_open((FIL*)&fp, record_number_filename, FA_WRITE) == FR_OK)
		{
			uint32_t writedsize;
			uint8_t data = face_id;
			f_write((FIL*)&fp, (uint8_t*)&data, 1, (UINT*)&writedsize);
			f_close((FIL*)&fp);
		}
	}
	return 0;
}

static int load_face_number(uint16_t *num)
{
	if (f_open((FIL*)&fp, record_number_filename, FA_READ) == FR_OK)
	{
		uint32_t readsize;
		uint8_t data;
		if (f_read((FIL*)&fp, (uint8_t*)&data, 1, (UINT*)&readsize) == FR_OK)
		{
			*num = data;
		}
		f_close((FIL*)&fp);
	}
	return 0;
}

#ifdef __cplusplus
}
#endif


void cv_face_init(uint16_t *recorded_face_number, uint8_t algorithm)
{
	train_model_ready = 0;
	*recorded_face_number = 0;

	if (algorithm & FRA_LBPH)
	{
		LBPHRecog = createLBPHFaceRecognizer(1,8,8,8,FACE_RECOGNITION_THRESHOLD_L); //createLBPHFaceRecognizer();
		if (f_open((FIL*)&fp, model_filename_lbph, FA_READ) == FR_OK)
		{
			f_close((FIL*)&fp);
			LBPHRecog->load(model_filename_lbph);
			train_model_ready = 1;
		}
	}
//	if (algorithm & FRA_EIGEN)
//	{
//		EigenRecog = createEigenFaceRecognizer(0, FACE_RECOGNITION_THRESHOLD_E);
//		if (f_open((FIL*)&fp, model_filename_eigen, FA_READ) == FR_OK)
//		{
//			f_close((FIL*)&fp);
//			EigenRecog->load(model_filename_eigen);
//			train_model_ready = 1;
//		}
//	}
//	if (algorithm & FRA_FISHER)
//	{
//		FisherRecog = createFisherFaceRecognizer(0, FACE_RECOGNITION_THRESHOLD_L);
//		if (f_open((FIL*)&fp, model_filename_fisher, FA_READ) == FR_OK)
//		{
//			f_close((FIL*)&fp);
//			FisherRecog->load(model_filename_fisher);
//			train_model_ready = 1;
//		}
//	}
	algorithmSelected = algorithm;

	load_face_number(recorded_face_number);
	total_faces = *recorded_face_number;

}

int cv_face_recognition(uint8_t* src, uint16_t width, uint16_t height, int * presult)
{
	int predictedLabel0 = -1;
	int predictedLabel1 = -1;
	int predictedLabel2 = -1;
	double confidence = 0.0;
	Mat img;

	if (train_model_ready == 0)
	{
		*presult = -1;
		return -1;
	}

	img = Mat(height,width,CV_8UC1,src);

	if (algorithmSelected & FRA_LBPH)
	{
		LBPHRecog->predict(img, predictedLabel0, confidence);
		*presult = predictedLabel0;
	}
//	if (algorithmSelected & FRA_EIGEN)
//	{
//		EigenRecog->predict(img, predictedLabel1, confidence);
//		presult[1] = predictedLabel1;
//	}
//	if (algorithmSelected & FRA_FISHER)
//	{
//		FisherRecog->predict(img, predictedLabel2, confidence);
//		presult[2] = predictedLabel2;
//	}

	return 0;
}

int cv_face_train(uint8_t* src, uint16_t width, uint16_t height, uint16_t face_label)
{
	int result;
	Mat img;
	FIL* fp;

	img = Mat(height,width,CV_8UC1,src);
	trainPics.push_back(img);
	labels.push_back(face_label);

	if (train_model_ready == 0)
	{
		if (algorithmSelected & FRA_LBPH)
		{
			LBPHRecog->train(trainPics, labels);
		}
//		if (algorithmSelected & FRA_EIGEN)
//		{
//			EigenRecog->train(trainPics, labels);
//		}
//		if (algorithmSelected & FRA_FISHER)
//		{
//			FisherRecog->train(trainPics, labels);
//		}
	}
	else
	{
		if (algorithmSelected & FRA_LBPH)
		{
			LBPHRecog->update(trainPics, labels);
		}
	}

	trainPics.clear();
	labels.clear();

	train_model_ready = 1;

	return 0;
}

int cv_face_train_terminate(uint32_t *pn)
{
	if (train_model_ready == 0)
		return 0;


	if (algorithmSelected & FRA_LBPH)
	{
		LBPHRecog->save(model_filename_lbph);
	}
//	if (algorithmSelected & FRA_EIGEN)
//	{
//		EigenRecog->save(model_filename_eigen);
//	}
//	if (algorithmSelected & FRA_FISHER)
//	{
//		FisherRecog->save(model_filename_fisher);
//	}

	*pn = *pn + 1;
	save_face_number(*pn);

	return 0;
}
#endif




